' -------------------------------------------- '
'                                              '
'     "Particle flow perfection"  by dafhi     '
'                                              '
'  written for Basic Gaming issue #5           '
'                        by Lachie Dazdarian   '
'                                              '
' -------------------------------------------- '

' A particles demo to showcase flow rate calculation
'to prevent sputtering

'screen dimensions
dim shared as Integer  SCR_W = 480
dim shared as Integer  SCR_H = 360

' middle of the screen
dim shared as single   midx: midx = (SCR_W - 1) / 2
dim shared as single   midy: midy = (SCR_H - 1) / 2

'adjust the action for different power cpu
dim shared as double   time_delta
dim shared as double   time_old
dim shared as double   time_new

'The Greek symbol is well-known for working with angles
Const                  TwoPi = 8 * Atn(1)

'basic particle
Type Particle
    as single          x
    as single          y
	As Single          dx
	As Single          dy
	As Single          age
	As Single          lifetime
End Type

/' more advanced particles could have density, size, shape, or whatever you could want
in your particle system.

there are also many different ways one could organize particles.
here is the method I chose - a second udt: '/

Type ParticleGroup
    As Single          lifetime_average  'average particle lifetime
    As Single          lifetime_variance 'average + variance = new particle lifetime
    
                                         'Actually, so lifetime will hover
                                         'the true average, lifetime is really:
                                         'average + variance * (Rnd - 0.5)
    
    As UInteger        AverageMax        'once all particles are released, the nr.
                                         'will stay close to this value
                                         
    As UInteger        StackPtr          'for recycling of particles
    As UInteger        ArrayUB           'the upper bound of the array .. [0 to arraysize - 1]
    
    'FreeBasic lets you call subroutines associated with the UDT!
    'a descriptor .. you will see the real instructions for Init() after "End Type"
    Declare Sub        Init(pArray() As Particle, _
                            ByVal avg_lifetime_ As Single = 2.5, _
                            ByVal lifetime_variance_ As Single = 1.0, _
                            ByVal Count_ As UInteger = 50)
    
    'here are two more definitions, which will go inside the "game loop"
    Declare Sub        EmitPerFrame(pArray() As Particle, ByVal rate_0_to_1 as single = 1.0, ByVal posx_ as single = 0, ByVal posy_ as single = 0)
    Declare Function   Render(pArray() as Particle) as integer
End Type

    'the definition has completed.  it is like a list of ingredients for a recipe.
    'to bake the actual cake, we need to describe what to do with the ingredients!

Sub ParticleGroup.Init(pArray() As Particle, _
                       ByVal avg_lifetime_ As Single, _
                       ByVal lifetime_variance_ As Single, _
                       ByVal Count_ As UInteger)
                       
	If Count_ > 150000 Then Exit Sub 'safety to prevent creation of many many particles.

    'make this larger if you absolutely must have true average for your super
    'important scientific experiment
    Dim as Integer AllowExtra = Count_ * 0.1
    
    'AverageMax is a member of ParticleGroup
	AverageMax = Count_
    
    'ArrayUB is a member of ParticleGroup
    ArrayUB = AverageMax - 1 + AllowExtra

    'pArray() is our array of particles, but not a member of ParticleGroup
    ReDim pArray(ArrayUB)
    
    'StackPtr is a member of ParticleGroup
	StackPtr = ArrayUB + 1

    'these are members of ParticleGroup
    lifetime_average = avg_lifetime_
	lifetime_variance = lifetime_variance_
    
End Sub

    'end of that one!  that sub will be used before the start of game loop.
    'now we will describe the process of our subs for inside the loop!

Sub ParticleGroup.EmitPerFrame(pArray() As Particle, ByVal rate_0_to_1 as Single, ByVal posx_ as single, ByVal posy_ as single)

    'it is common to have safety protocol to prevent noob error
    if AverageMax < 1 then exit sub
    
    'rate_0_to_1 can be used to control the flow like a valve
    dim as integer Nr_ = rate_0_to_1 * time_delta * AverageMax / lifetime_average
    
    /' Here is a diagram of my particles recycling system
    
    [0] [1] [2] [3] [4] [5] [6] [7] [8] [9]
	
    [   recycle   ] [       active        ]
    [ end   start ] [ StackPtr .. ArrayUB ] '/
    
	Dim As Integer recycle_start = StackPtr - 1
	Dim As Integer recycle_end = recycle_start - (Nr_ - 1)
    
    'particle characteristics
    dim as single velocity
    dim as single angle

    'The particles index cannot go below zero
    If recycle_end < 0 Then recycle_end = 0
    
    For I As Integer = recycle_start To recycle_end Step -1
        
        'shoot away at any angle
        angle = Rnd * TwoPi
        
        'with varying speeds
        velocity = 10 * (1.0 + Rnd)
        
        'pArray() is a shared array.  Not a type member of ParticleGroup.
        With pArray(I)
            
            'animation frame is an instantaneous moment in time,
            'and true object information can be any value in between,
            'not 'digital' like animation frame
            .age = Rnd * time_delta
            
            'I could also create 'not digital' data for particle origin,
            'based on some function, but this effect looks nice
            .x = posx_
            .y = posy_
            
            'movement
            .dx = velocity * Cos(angle)
            .dy = velocity * Sin(angle)
    
            .lifetime = lifetime_average + lifetime_variance * (Rnd - 0.5)
            
        End With
        
    Next
    
    Nr_ = recycle_start - recycle_end + 1
    
    StackPtr -= Nr_ '' adjust StackPtr so Render() will use new particles
    
End Sub 'ParticleGroup.EmitThisFrame


    'and here is Render sub

Function ParticleGroup.Render(pArray() as Particle) as integer

    'a simple check if ParticleGroup was initialized
    if AverageMax < 1 then exit Function
    
    'I use this to print info on active particles
    dim as integer active_
    
    /'If you remember: Array positions
    [0] [1] [2] [3] [4] [5] [6] [7] [8] [9]
    [  inactive   ] [ StackPtr to ArrayUB ] '/

    'these are active particles
    For I as integer = StackPtr to ArrayUB
        
        With pArray(I) 'now we are dealing with a specific active particle.
            
            if .age < .lifetime then
                
                PSet ( .x + .age * .dx, _
                       .y + .age * .dy _
                     ), &HFFFFFFFF

                .age += time_delta
                
                active_ += 1
                
            Else /' particle has met its demise

                    array before:
                    [0] [0] [0] [1] [1] [1] [0] <- spent particle
                                 #   SWAP    #
                    array after:
                    [0] [0] [0] [0] [1] [1] [1]
                                    [ StackPtr + 1 ]
                '/
                
                Swap pArray(StackPtr), pArray(I)
                
                StackPtr += 1
                
            End if
            
        End With
        
    Next I
    
    return active_ ' put the value into the function
    
End Function

    'we're almost done!  the rest is a piece of cake.
Dim As Particle         Particles()
Dim As ParticleGroup    PG

    'initialize our graphics screen
ScreenRes SCR_W, SCR_H, 32

    'bring in code from freebasic "library" to handle special keyboard functions
#Include "fbgfx.bi"
Using FB
Dim As EVENT e

    'begin timing
time_old = Timer

    'some variables to help programmer remember his vision
dim as single           life_avg = 1.0
dim as single           life_vari = 0.5
dim as integer          Nr = 150

    'initialize!
PG.Init Particles(), life_avg, life_vari, Nr

    'here is the game loop
Do
    
    'timing algorithm
    time_new = Timer
    time_delta = time_new - time_old
    time_old = time_new
    
    PG.EmitPerFrame Particles(), 1.0, _
        midx + 250 * (Rnd - 0.5), _
        midy + 250 * (Rnd - 0.5)

    ScreenLock ' this is freebasic to eliminate "tearing" artifact.
               ' you must have ScreenUnlock when you are done drawing.
               
    Cls        ' clears the screen
    
	'' PG.Render does two things:  draw and return information
    Print "Particles: " & PG.Render( Particles() )
    
    ScreenUnlock ' Done drawing!
    
    ' This lets us exit the program.
    If (ScreenEvent(@e)) Then
		If e.scancode = SC_ESCAPE Then
			Exit Do
		End If
    End If
    
    sleep 30 ''30 ms is slightly higher than 25 fps
    
Loop
